#!ruby
# Delphi Parser
# author: YOSHIDA Kazuhiro (moriq)
# mailto: moriq@moriq.com

require 'dp/tokenizer'
require 'dp/structure'

class Parser

private

def list_loop(delimiter)
  loop do
    yield
    case @t.dc_token
    when delimiter
      @t.next_token
    else
      break
    end
  end
end

def list_loop_end
  case $!
  when TokenError
    #
  else
    @d.errputs $!, $!.backtrace
  end
end

def ident_list
  ret = []
  list_loop(',') { ret.push @t.check(:ident) }
  ret
end

def class_herit
  @t.check '('
  ret = ident_list
  @t.check ')'
  ret
end

def variant
  ret = Delphi::Variant.new
  begin
    list_loop(','){
      expr = []
      expr.push @t.check(:ident)
      while true
        case @t.dc_token
        when 'or'
          expr.push @t.next_token
          expr.push @t.check(:ident)
        else
          break
        end
      end
      ret.expr_list.push expr
    }
  rescue
    list_loop_end
  end
  case @t.dc_token
  when ':'
    @t.next_token
    @t.check '('
    ret.field_list = field_list
    @t.check ')'
  end
  ret
end

def variant_part
  ret = Delphi::VariantPart.new
  @t.check 'case'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when ':'
    @t.next_token
    ret.type = type_ident
  end
  @t.check 'of'
  list_loop(';'){ ret.variant_list.push variant }
  ret
end

def fixed_part
  ret = Delphi::Field.new
  ret.ident_list = ident_list
  @t.check(':')
  ret.type = normal_type
  ret
end

def field_list
  @d.outputs "### field_list"
  ret = Delphi::FieldList.new
  begin
    list_loop(';'){ ret.fixed_list.push fixed_part }
  rescue
    list_loop_end
  end
  case @t.dc_token
  when 'case'
    ret.variant_part = variant_part
  end
  ret
end

def param
  ret = Delphi::Param.new
  ret.ident_list = ident_list
  case @t.dc_token
  when ':'
    @t.next_token
    ret.type =
    case @t.dc_token
    when 'array'
      ret_type = []
      ret_type.push @t.next_token
      ret_type.push @t.check('of')
      ret_type.push \
      case @t.dc_token
      when 'const'
        @t.next_token
      else
        ord_type
      end
      ret_type
    when 'string'
      @t.next_token
    when 'file'
      @t.next_token
    else
      normal_type
    end
    case @t.dc_token
    when '='
      @t.next_token
      ret.value = const_expr
    end
  end
  ret
end

def formal_param
  case @t.dc_token
  when 'var', 'const', 'out'
    @t.next_token
  end
  param
end

def formal_param_list
  ret = []
  @t.check '('
  begin
    list_loop(';'){ ret.push formal_param }
  rescue
    list_loop_end
  end
  @t.check ')'
  ret
end

def directive
  @d.outputs "### directive"
  delim = nil
  case @t.dc_token
  when ';'
    delim = @t.next_token
  end
  ret = []
loop do
  case @t.dc_token
  when 'cdecl', 'register', 'dynamic', 'virtual', 'export', 'near', 'far', 'override', 'overload', 'pascal', 'reintroduce', 'safecall', 'stdcall', 'varargs', 'local', 'abstract', 'assembler'
    ret.push @t.next_token
# when 'external'
#   ret.push @t.next_token
#   const_expr
#   if @t.dc_token == 'name'
#     ret.push @t.next_token
#     const_expr
#   end
# when , 'forward'
#   ret.push @t.next_token
  else
    break
  end
  case @t.dc_token
  when ';'
    delim = @t.next_token
  end
end
    @t.put_token(delim) if delim
  ret
end

def portability_directive
  @d.outputs "### portability_directive"
  delim = nil
  case @t.dc_token
  when ';'
    delim = @t.next_token
  end
  case @t.dc_token
  when 'deprecated', 'library', 'platform'
    @t.next_token
  else
    @t.put_token(delim) if delim
  end
end

def procedure_header
  ret = Delphi::Procedure.new
  @t.check 'procedure'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '.'
    @t.next_token
    @t.check(:ident)
    @t.check '='
    @t.check(:ident)  # assign interface method
    return
  when '('
    ret.param_list = formal_param_list
  end
  delim = nil
  case @t.dc_token
  when ';'
    delim = @t.next_token
  end
  case @t.dc_token
  when 'message'
    @t.next_token
    const_expr
  else
    @t.put_token(delim) if delim
  end
  ret.directive = directive
  portability_directive
  ret
end

def procedure_header_decl
  ret = Delphi::Procedure.new
  @t.check 'procedure'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '.'
    @t.next_token
    ret.ident = @t.check(:ident)
  end
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  ret.directive = directive
  portability_directive
  ret
end

def function_header
  ret = Delphi::Function.new
  @t.check 'function'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '.'
    @t.next_token
    @t.check(:ident)
    @t.check '='
    @t.check(:ident)  # assign interface method
    return
  when '('
    ret.param_list = formal_param_list
  end
  @t.check ':'
  ret.type = normal_type
  ret.directive = directive
  portability_directive
  ret
end

def function_header_decl
  ret = Delphi::Function.new
  @t.check 'function'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '.'
    @t.next_token
    ret.ident = @t.check(:ident)
  end
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  case @t.dc_token
  when ':'
    @t.next_token
    ret.type = normal_type
  end
  ret.directive = directive
  portability_directive
  ret
end

def constructor_header
  ret = Delphi::Constructor.new
  @t.check 'constructor'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  ret.directive = directive
  portability_directive
  ret
end

def constructor_header_decl
  ret = Delphi::Constructor.new
  @t.check 'constructor'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '.'
    @t.next_token
    ret.ident = @t.check(:ident)
  end
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  ret.directive = directive
  portability_directive
  ret
end

def destructor_header
  ret = Delphi::Destructor.new
  @t.check 'destructor'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  ret.directive = directive
  portability_directive
  ret
end

def destructor_header_decl
  ret = Delphi::Destructor.new
  @t.check 'destructor'
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '.'
    @t.next_token
    ret.ident = @t.check(:ident)
  end
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  ret.directive = directive
  portability_directive
  ret
end

def method_header
  @d.outputs "### method_header"
  case @t.dc_token
  when 'class'
    @t.next_token
    # class method
  end
  case @t.dc_token
  when 'procedure'
    procedure_header
  when 'function'
    function_header
  when 'constructor'
    constructor_header
  when 'destructor'
    destructor_header
  end
end

def prop_param_list
  @t.check '['
  list_loop(';') do
    case @t.dc_token
    when 'var', 'const', 'out'
      @t.next_token
    end
    ident_list
    @t.check ':'
    type_ident
  end
  @t.check ']'
end

# def prop_interface
#   case @t.dc_token
#   when '['
#     prop_param_list
#   end
#   @t.check ':'
#   normal_type
# end

def prop_list
  @d.outputs "### prop_list"
  ret = Delphi::Property.new
  @t.check 'property'
  ret.ident = @t.check(:ident)

  case @t.dc_token
  when '[', ':'
  # ret.interface = prop_interface
    case @t.dc_token
    when '['
      ret.param_list = prop_param_list
    end
    @t.check ':'
    ret.type = normal_type
  end

  case @t.dc_token
  when 'index'
    @t.next_token
    const_expr
  end
  case @t.dc_token
  when 'read'
    @t.next_token
    type_ident
  end
  case @t.dc_token
  when 'write'
    @t.next_token
    type_ident
  end
  case @t.dc_token
  when 'stored'
    @t.next_token
    type_ident
  end
  case @t.dc_token
  when 'default'
    @t.next_token
    const_expr
  when 'nodefault'
    @t.next_token
  end
  case @t.dc_token
  when 'implements'
    @t.next_token
    list_loop(',') { @t.check(:ident) }
  end
  case @t.dc_token
  when 'readonly', 'writeonly'
    @t.next_token
  end
  case @t.dc_token
  when 'dispid'
    @t.next_token
    const_expr
  end
  delim = nil
  case @t.dc_token
  when ';'
    delim = @t.next_token
  end
  case @t.dc_token
  when 'default'
    @t.next_token
  else
    @t.put_token(delim) if delim
  end
  ret
end

def class_visibility
  case @t.dc_token
  when 'private', 'protected', 'public', 'published'
    @t.next_token
# else
#   raise
  end
end

def obj_field_list
  @d.outputs "### obj_field_list"
  ret = Delphi::ObjFieldList.new
  list_loop(';') do
    ret.ident_list = ident_list
    @t.check(':')
    ret.type = normal_type
  end
  ret
end

def class_field_list
  @d.outputs "### class_field_list"
  ret = []
  begin
    list_loop(';') do
#     class_visibility
      ret.concat obj_field_list
    end
  rescue
    list_loop_end
  end
  ret
end

def method_list
  @d.outputs "### method_list"
  ret = method_header
  case @t.dc_token
  when 'dispid'
    @t.next_token
    const_expr
  end
  ret
end

def class_method_list
  @d.outputs "### class_method_list"
  ret = []
  begin
    list_loop(';') do
#     class_visibility
      ret.push method_list
    end
  rescue
    list_loop_end
  end
  ret
end

def class_prop_list
  @d.outputs "### class_prop_list"
  ret = []
  begin
    list_loop(';') do
#     class_visibility
      ret.push prop_list
    end
  rescue
    list_loop_end
  end
  ret
end

def class_type
  @t.check 'class'
  case @t.dc_token
  when 'of'
    @t.next_token
    cref = Delphi::ClassRef.new
    cref.base = type_ident
    return cref
  end
  dc = nil
  case @t.dc_token
  when ';'
    dc = Delphi::Class.new
    return dc
  when '('
    herit = class_herit
    dc = Delphi::Class.new(herit)
    case @t.dc_token
    when ';'
      return dc
    when 'end'
      @t.next_token
      return dc
    end
  end
  @d.outputs "### enter class_type " << @t.dc_token
  visib = class_visibility || 'published'
  loop do
    case @t.dc_token
    when 'private', 'protected', 'public', 'published'
      visib = class_visibility
    when 'procedure', 'function', 'constructor', 'destructor'
      dc.add_method_list visib, class_method_list
    when 'property'
      dc.add_prop_list visib, class_prop_list
    when 'end'
      break
    else
      dc.add_field_list visib, class_field_list
    end
  end
  @t.check 'end'
  @d.outputs "### leave class_type " << @t.dc_token
  dc
end

def interface_herit
  @t.check '('
  ret = ident_list
  @t.check ')'
  ret
end

def interface_type
  @t.check 'interface'
  case @t.dc_token
  when ';'
    return
  when '('
    interface_herit
    case @t.dc_token
    when ';'
      return
    when 'end'
      @t.next_token
      return
    end
  end
  case @t.dc_token
  when '['
    @t.next_token
    @t.check(:ident)
    @t.check ']'
  end
  loop do
    case @t.dc_token
    when 'private', 'protected', 'public', 'published'
      class_visibility
    when 'procedure', 'function', 'constructor', 'destructor'
      class_method_list
    when 'property'
      class_prop_list
    when 'end'
      break
    end
  end
  @t.check 'end'
end

def dispinterface_type
  @t.check 'dispinterface'
  case @t.dc_token
  when ';'
    return
  when '('
    interface_herit
    case @t.dc_token
    when ';'
      return
    when 'end'
      @t.next_token
      return
    end
  end
  case @t.dc_token
  when '['
    @t.next_token
    @t.check(:ident)
    @t.check ']'
  end
  loop do
    case @t.dc_token
    when 'private', 'protected', 'public', 'published'
      class_visibility
    when 'procedure', 'function', 'constructor', 'destructor'
      class_method_list
    when 'property'
      class_prop_list
    when 'end'
      break
    end
  end
  @t.check 'end'
end

def pointer_type
  ret = Delphi::PointerType.new
  loop do
    case @t.dc_token
    when '^'  # pointer
      ret.ref.push @t.next_token
    else
      break
    end
  end
  ret.type = @t.check(:ident)
end

# def ord_ident
#   @t.check(:ident)
# end

# def subrange_type
#   const_expr
#   @t.check '..'
#   const_expr
# end

def ord_type
  @d.outputs "### ord_type"
  case @t.dc_token
  when '('
    enum_type
  else
    # ord_ident
    # subrange_type
    ord_expr = const_expr
    case @t.dc_token
    when '..'
      @t.next_token
      ret = Delphi::SubRangeType.new
      ret.begin_expr = ord_expr
      ret.end_expr = const_expr
    else
      ret = Delphi::OrdType.new
      ret.expr = ord_expr
    end
    ret
  end
end

def array_type
  ret = Delphi::ArrayType.new
  loop do
    case @t.dc_token
    when 'array'
      @t.next_token
      range = []
      case @t.dc_token
      when '['
        @t.next_token
        list_loop(','){ range.push ord_type }
        @t.check ']'
      end
      @t.check 'of'
      ret.range.push range
    else
      break
    end
  end
  ret.type =
  case @t.dc_token
  when 'const'
    @t.next_token
  when 'procedure', 'function'
    proc_type
  else
    ord_type
  end
  ret
end

def string_type
  ret = Delphi::StringType.new
  @t.check 'string'
  case @t.dc_token
  when '['
    @t.next_token
    ret.type = ord_type
    @t.check ']'
  end
  ret
end

def enum_type_item
  ret = Delphi::EnumTypeItem.new
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when '='
    @t.next_token
    ret.value = const_expr
  end
  ret
end

def enum_type
  ret = Delphi::EnumType.new
  @t.check '('
  list_loop(','){ ret.items.push enum_type_item }
  @t.check ')'
  ret
end

def record_field_list
  @d.outputs "### record_field_list"
# ret = []
# begin
#   list_loop(';'){ ret.push field_list }
# rescue
#   list_loop_end
# end
  ret = field_list
  ret
end

def record_type
  @t.check 'record'
  ret = record_field_list
  @t.check 'end'
  ret
end

def file_type
  ret = Delphi::FileType.new
  @t.check 'file'
  case @t.dc_token
  when 'of'
    @t.next_token
    ret.type = type_ident
  end
  ret
end

def set_type
  ret = Delphi::SetType.new
  @t.check 'set'
  @t.check 'of'
  ret.type = ord_type
  ret
end

def procedure_type
  ret = Delphi::ProcedureType.new
  @t.check 'procedure'
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  ret
end

def function_type
  ret = Delphi::FunctionType.new
  @t.check 'function'
  case @t.dc_token
  when '('
    ret.param_list = formal_param_list
  end
  @t.check ':'
  ret.type = normal_type
  ret
end

def proc_type
  @d.outputs "### proc_type"
  ret =
  case @t.dc_token
  when 'procedure'
    procedure_type
  when 'function'
    function_type
  end
  case @t.dc_token
  when 'of'
    @t.next_token
    @t.check 'object'
  end
  ret.directive = directive
  portability_directive
  ret
end

def normal_type
  @d.outputs "### normal_type"
  ret = 
  case @t.dc_token
  when 'type'
    @t.next_token
    normal_type
  when 'procedure', 'function'
    proc_type
  when 'class'
    class_type
  when 'interface'
    interface_type
  when 'dispinterface'
    dispinterface_type
  when 'packed'
    @t.next_token
    normal_type
  when 'record'
    record_type
  when 'file'
    file_type
  when 'set'
    set_type
  when '('
    enum_type
  when '^'
    pointer_type
  when 'array'
    array_type
  when 'string'
    string_type
  else
    ord_type
  end
  portability_directive
  ret
end

def type_ident
  @d.outputs "### type_ident"
  ret = []
  list_loop('.'){ ret.push @t.check(:ident) }
  ret
end

def factor
  @d.outputs "### factor"
  ret = []
  if @t.dc_token == '^'
    ret.push @t.next_token
    ret.push @t.next_token # ctrl char ???
    return ret
  end
  if /^['$#]/ =~ @t.dc_token
    while /^['$#]/ =~ @t.dc_token
      @d.outputs "### string #{@t.dc_token}"
      ret.push @t.next_token
    end
    return ret
  end
  case @t.dc_token
  when '@', 'not', 'inherited'# '+', '-'
    ret.push @t.next_token
  end
  case @t.dc_token
  when '@'
    ret.push @t.next_token
  end
  if @t.ident?(@t.dc_token)
    ret.push variable
    return ret
  end
  case @t.dc_token
  when '('
    @t.next_token
    list_loop(';'){
      ret.concat expr_list
      case @t.dc_token
      when ':'
        @t.next_token
        ret.push expr
      end
    }
    @t.check ')'
    case @t.dc_token
    when '^'
      ret.push @t.next_token
    end
    case @t.dc_token
    when '.'
      ret.push @t.next_token
      ret.push variable
    end
    return ret
  when '['
    ## SetConstructor
    @t.next_token
    ret.concat expr_list
    @t.check ']'
    return ret
  end
# @d.outputs "nil factor #{@t.dc_token}"
# raise TokenError.new("nil factor #{@t.dc_token}")
end

def term
  @d.outputs "### term"
  ret = []
  ret.push factor
  loop do
    case @t.dc_token
    when '*', '/', 'div', 'mod', 'and', 'shl', 'shr' # mulop
      ret.push @t.next_token
    else
      return ret
    end
    ret.push factor
  end
  ret
end

def simple_expr
  @d.outputs "### simple_expr"
  ret = []
  case @t.dc_token
  when '+', '-'
    ret.push @t.next_token
  end
  ret.push term
  loop do
    case @t.dc_token
    when '+', '-', 'or', 'xor' # addop
      ret.push @t.next_token
    else
      return ret
    end
    ret.push term
  end
  ret
end

def expr_list
  ret = []
  list_loop(','){ ret.push expr }
  ret
rescue
  list_loop_end
  ret
end

def expr
  ret = []
  ret.push simple_expr
  case @t.dc_token
  when '..'
    ret.push @t.next_token
    ret.push simple_expr
    return ret
  when '=', '<>', '<', '>', '<=', '>=', 'in', 'is', 'as' # relop
    ret.push @t.next_token
    ret.push simple_expr
    return ret
  when ':'
    ret.push @t.next_token
    ret.push simple_expr
    return ret
  end
  ret
end

def const_expr
  simple_expr
end

def uses
  @t.check 'uses'
  ret = []
  loop do
    ret.push @t.check(:ident)
    case @t.dc_token
    when ','
      @t.next_token
    when ';'
      @t.next_token
      break
    else
      raise "uses: unknown #{@t.dc_token}"
    end
  end
  ret
end

def const_decl
  ret = Delphi::Const.new
  ret.ident = @t.check(:ident)
  case @t.dc_token
  when ':'
    @t.next_token
    ret.type = normal_type
  end
  @d.outputs "### @t.check = in const_decl"
  @t.check '='
  ret.value = const_expr
  # for Indy/IdCoder3To4.pas
  loop do
    case @t.dc_token
    when '='
      @d.outputs "### @t.check = in const_decl too"
      @t.next_token
      ret.value = const_expr
    else
      break
    end
  end
  portability_directive
  ret
end

def const_decl_section
  @t.check 'const'
  ret = []
  begin
    list_loop(';'){ ret.push const_decl }
  rescue
    list_loop_end
  end
  ret
end

def resourcestring_decl_section
  @t.check 'resourcestring'
  ret = []
  begin
    list_loop(';'){ ret.push const_decl }
  rescue
    list_loop_end
  end
  ret
end

def type_decl
  ident = @t.check(:ident)
  @t.check '='
  value = normal_type
  direc = directive
  portability_directive
  ret = Delphi::Type.new(ident, value, direc)
  @on_type_decl.call(ret) if @on_type_decl
  ret
end

def type_decl_section
  @t.check 'type'
  ret = []
  begin
    list_loop(';'){ ret.push type_decl }
  rescue
    list_loop_end
  end
  ret
end

def var_decl
  @d.outputs "### var_decl"
  ret = Delphi::Var.new
  ret.ident_list = ident_list
  @t.check ':'
  ret.type = normal_type
  ret.directive = directive
  case @t.dc_token
  when 'absolute'
    @t.next_token
    ret.absolute = @t.check(:ident)
  when '='
    @t.next_token
    @d.outputs "### = in var_decl"
    ret.value = const_expr
  end
  portability_directive
  ret
end

def var_decl_section
  @t.check 'var'
  ret = []
  begin
    list_loop(';'){ ret.push var_decl }
  rescue
    list_loop_end
  end
  ret
end

def threadvar_decl_section
  @t.check 'threadvar'
  ret = []
  begin
    list_loop(';'){ ret.push var_decl }
  rescue
    list_loop_end
  end
  ret
end

def procedure_decl
  @d.outputs "### procedure_decl"
  ret = procedure_header_decl
  @t.check ';'
  proc_block
  ret
end

def function_decl
  @d.outputs "### function_decl"
  ret = function_header_decl
  @t.check ';'
  proc_block
  ret
end

def constructor_decl
  @d.outputs "### constructor_decl"
  ret = constructor_header_decl
  @t.check ';'
  proc_block
  ret
end

def destructor_decl
  @d.outputs "### destructor_decl"
  ret = destructor_header_decl
  @t.check ';'
  proc_block
  ret
end

def proc_block
  @d.outputs "### proc_block"
  case @t.dc_token
  when 'external'
    @d.outputs "### external"
    @t.next_token
    if @t.dc_token != ';'
      const_expr
      if @t.dc_token == 'name'
        @d.outputs "### name"
        @t.next_token
        const_expr
      end
    end
    directive
    if @t.dc_token == ';'
      @t.next_token
    end
  when 'forward'
    @d.outputs "### forward"
    @t.next_token
    @t.check ';'
  else
    block
  end
end

def block
  @d.outputs "### block"
  decl_sect_list
  case @t.dc_token
  when 'begin'
    compound_stmt
  when 'asm'
    asm_stmt
  end
end

def compound_stmt
  @d.outputs "### compound_stmt"
  @t.check 'begin'
  ret = stmt_list
  @t.check 'end'
  ret
end

def stmt_list
  ret = []
  begin
    list_loop(';'){ ret.push stmt }
  rescue
    list_loop_end
  end
  ret
end

def stmt
  @d.outputs "### stmt"
# :empty,
# 'begin', 'goto', 'if', 'case', 'repeat', 'while', 'for', 'with',
# 'try', 'asm', 'raise', 'inherited', '@', '(', :ident
  case @t.dc_token
  when 'begin'
    compound_stmt
  when 'goto'
    goto_stmt
  when 'if'
    if_stmt
  when 'case'
    case_stmt
  when 'repeat'
    repeat_stmt
  when 'while'
    while_stmt
  when 'for'
    for_stmt
  when 'with'
    with_stmt
  when 'try'
    try_statement # /* purely Delphi stuff ! */
  when 'asm'
    asm_stmt
  when 'raise'
    raise_stmt
  else
    # 'inherited', '@', '(', :ident
    token = @t.dc_token
    token = :ident if @t.ident?(token)
    if ['inherited', '@', '(', :ident].include?(token)
      general_proc_call_or_assignment
    # assignment
    end
  end
end

def assignment
  variable
  @t.check(':=')
  expr
end

def goto_stmt
  @t.check('goto')
  @t.check(:ident) # label
end

def general_proc_call_or_assignment
  @d.outputs "### general_proc_call_or_assignment"
# 'inherited', '@', '(', :ident
  case @t.dc_token
  when 'inherited'
    @t.next_token
  end
  token = @t.dc_token
  token = :ident if @t.ident?(token)
  if ['@', '(', :ident].include?(token)
    variable
  end
  case @t.dc_token
  when ':='
    @d.outputs "### assignment"
    @t.next_token
    expr
  when '.'
    @t.next_token
    proc_call
  end
end

def general_proc_call
# 'inherited', '@', '(', :ident
  case @t.dc_token
  when 'inherited'
    @t.next_token
  end
  token = @t.dc_token
  token = :ident if @t.ident?(token)
  if ['@', '(', :ident].include?(token)
    proc_call
  end
end

def proc_call
  list_loop('.') do
    variable # instead if proc_ident param_list
  end
rescue
  list_loop_end
end

def if_stmt
  @t.check('if')
  expr
  @t.check('then')
  stmt
  case @t.dc_token
  when 'else'
    @t.check('else')
    stmt
  end
end

def case_stmt
  @t.check('case')
  expr
  @t.check('of')
  case_list
  case @t.dc_token
  when 'else'
    @t.check('else')
    stmt_list
    if @t.dc_token == ';' then @t.next_token end
  end
  @t.check('end')
end

def case_list
  list_loop(';'){
    if @t.dc_token == 'else' then break end
    case_label_list
    @t.check(':')
    stmt
  }
rescue
  list_loop_end
end

def case_label_list
  list_loop(','){ case_label }
rescue
  list_loop_end
end

def case_label
  list_loop('..'){ const_expr }
rescue
  list_loop_end
end

def repeat_stmt
  @t.check('repeat')
  stmt_list
  @t.check('until')
  expr
end

def while_stmt
  @t.check('while')
  expr
  @t.check('do')
  stmt
end

def for_stmt
  @t.check('for')
  @t.check(:ident)
  @t.check(':=')
  expr
  case @t.dc_token
  when 'to', 'downto'
    @t.next_token
  else
    raise TokenError
  end
  expr
  @t.check('do')
  stmt
end

def with_stmt
  @t.check('with')
# variable_list
  expr_list
  @t.check('do')
  stmt
end

def variable_list
  list_loop(','){ variable }
rescue
  list_loop_end
end

def try_statement
  @t.check('try')
  stmt_list
  case @t.dc_token
  when 'finally'
    @t.next_token
    stmt_list
    @t.check('end')
  when 'except'
    @t.next_token
    token = @t.dc_token
    token = :ident if token != 'on' && @t.ident?(token)
    case token
    when ';'
      stmt_list
    when 'begin', 'goto', 'if', 'case', 'repeat', 'while', 'for', 'with',
         'try', 'asm', 'raise', 'inherited', '@', '(', :ident
      stmt_list
    when 'on'
      general_exception_handlers_list
    end
    @t.check('end')
  else
    raise TokenError
  end
end

def general_exception_handlers_list
  @d.outputs "### general_exception_handlers_list"
  exception_handlers_list
  exception_else_clause
end

def exception_handlers_list
  list_loop(';'){
    if @t.dc_token == 'else' then break end
    exception_handler
  }
rescue
  list_loop_end
end

def exception_handler
  if @t.dc_token == 'on' then
    @t.next_token
    exception_selector
    @t.check('do')
    stmt
  end
end

def exception_selector
  @t.check(:ident)
  if @t.dc_token == ':' then
    @t.next_token
    @t.check(:ident)
    if @t.dc_token == '.' then
      @t.next_token
      @t.check(:ident)
    end
  end
end

def exception_else_clause
  if @t.dc_token == 'else' then
    @t.next_token
    stmt_list
  end
end

def asm_stmt
  @t.check('asm')
  loop do
    if @t.dc_token == 'end' then
      @t.next_token
      if @t.dc_token == ';' then
        break
      end
    end
    @t.next_token
  end
end

def raise_stmt
  @t.check('raise')
  expr_list
  if @t.dc_token == 'at' then
    @t.next_token
    const_expr
  end
end

def variable
  @d.outputs "### variable"
  ret = []
# '@', '(', :ident
  case @t.dc_token
  when '@'
    ret.push @t.next_token
    ret.push variable
    return ret
  when '('
    @t.next_token
    ret.concat expr_list
    @t.check ')'
  else
    ret.push @t.check(:ident)
  end
loop do
  case @t.dc_token
  when '('
    @t.next_token
    ret.concat expr_list
    @t.check(')')
  when '['
    @t.next_token
    ret.concat expr_list
    @t.check(']')
  when '^'
    ret.push @t.next_token
  when '.'
    ret.push @t.next_token
    ret.push variable
  else
    break
  end
end
  ret
end

def interface_decl
  @t.check('interface')
  di = Delphi::Interface.new
  case @t.dc_token
  when 'uses'
    # di.uses.concat 
    uses
  end
  loop do
    case @t.dc_token
    when 'type'
      # di.types.concat 
      type_decl_section
    when 'const'
      # di.consts.concat 
      const_decl_section
    when 'resourcestring'
      # di.resourcestrings.concat 
      resourcestring_decl_section
    when 'var'
      # di.vars.concat 
      var_decl_section
    when 'threadvar'
      # di.threadvars.concat 
      threadvar_decl_section
    when 'procedure'
      # di.procedures.push 
      procedure_header
      @t.check ';'
    when 'function'
      # di.procedures.push 
      function_header
      @t.check ';'
    else
      break
    end
  end
  di
end

def decl_sect_list
  while true do # loop do
    case @t.dc_token
    when 'type'
      type_decl_section
    when 'const'
      const_decl_section
    when 'resourcestring'
      resourcestring_decl_section
    when 'var'
      var_decl_section
    when 'threadvar'
      threadvar_decl_section
    when 'procedure'
      procedure_decl
    when 'function'
      function_decl
    when 'constructor'
      constructor_decl
    when 'destructor'
      destructor_decl
    when 'class'
      @t.next_token
      case @t.dc_token
      when 'procedure'
        procedure_decl
      when 'function'
        function_decl
      end
    else
      break
    end
    if @t.dc_token == ';' then @t.next_token end
  end
  nil
end

def implementation_decl
  @t.check('implementation')
  case @t.dc_token
  when 'uses'
    uses
  end
  decl_sect_list
end

def initialization_part
  if @t.dc_token == 'initialization' then
    @d.outputs "### initialization"
    @t.next_token
    stmt_list
  end
end

def finalization_part
  if @t.dc_token == 'finalization' then
    @d.outputs "### finalization"
    @t.next_token
    stmt_list
  end
end

def unit
  @t.check('unit')
  du = Delphi::Unit.new
  du.ident = @t.check(:ident)
  portability_directive
  @t.check(';')
  du.interface = interface_decl
  implementation_decl
  initialization_part
  finalization_part
  if @t.dc_token == 'begin' then
    block
  else
    @t.check('end')
  end
  @t.check('.')
  du
end

def initialize
  @t = Tokenizer.new
end

public

def load(fname)
  @filename = fname
  @t.load(fname)
end

attr_reader :filename

def parse
  unit
end

attr_accessor :on_type_decl

def deb=(d)
  @d = d
  @t.deb = d
end

require 'forwardable'
extend Forwardable
def_delegators(:@t, :source, :offset, :lineno)

def fileinfo
  "#{filename}(#{lineno}):"
end

end
